package chocopy.pa3;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.*;

import chocopy.common.analysis.SymbolTable;
import chocopy.common.analysis.AbstractNodeAnalyzer;
import chocopy.common.analysis.types.SymbolType;
import chocopy.common.analysis.types.ValueType;
import chocopy.common.astnodes.*;
import chocopy.common.codegen.*;

import static chocopy.common.codegen.RiscVBackend.Register.*;

/**
 * This is where the main implementation of PA3 will live.
 *
 * A large part of the functionality has already been implemented
 * in the base class, CodeGenBase. Make sure to read through that
 * class, since you will want to use many of its fields
 * and utility methods in this class when emitting code.
 *
 * Also read the PDF spec for details on what the base class does and
 * what APIs it exposes for its sub-class (this one). Of particular
 * importance is knowing what all the SymbolInfo classes contain.
 */

/* Recolic func manage logic:
IMPORTANT!!! NEVER DELETE THESE COMMENTS

emitPush(arg1)
emitPush(arg2)
emitCall("someFunc")
emitPop(null)
emitPop(null)

$someFunc:

-------- emitFunctionBegin ------
emitPush(RA) // save My RA
emitPush(FP) // save parent FP // that's a good fucking design
emitMv(FP, SP+2*WORD) // load My FP
emitJ($someFunc$saveReg)
$someFunc$saveRegContinue:
-------- emitFunctionBegin ------

some code...
using S1...
using S2...
calling others...
...
MV A0, returnValue // IMPORTANT!

-------- emitFunctionEnd ------
emitMv(SP, FP-4*WORD) // 2 WORD for FP and RA, and 2 WORD for register S1 and S2.
emitPop(S2)
emitPop(S1)
emitPop(FP) // restore parent's FP
emitPop(RA) // restore My RA
emitJR(RA) // return!
$someFunc$saveReg:
emitPush(S1)
emitPush(S2)
emitJ($someFunc$saveRegContinue)
-------- emitFunctionEnd ------
 */



public class CodeGenImpl extends CodeGenBase {
    private class BetterRsicVBackend {
        private final RiscVBackend backend;
        BetterRsicVBackend(RiscVBackend _backend) {
            backend = _backend;
        }
        public void emitNoop(String comment) {
            if(comment != null) {
                backend.emitMV(ZERO, ZERO, "Noop: " + comment);
            }
        }
        public void emitPush(RiscVBackend.Register reg, String comment) {
            backend.emitADDI(SP, SP, -1 * backend.getWordSize(), comment);
            backend.emitSW(reg, SP, 0, comment);
        }
        public void emitPop(RiscVBackend.Register reg, String comment) {
            if(reg != null)
                backend.emitLW(reg, SP, 0, comment);
            backend.emitADDI(SP, SP, backend.getWordSize(), comment);
        }
        public void emitCall(Label calledLabel, String comment) {
            // Arguments should be already pushed to stack.
            backend.emitJAL(calledLabel, comment);
        }
        public void emitFunctionBegin(String funcName, String comment) {
            emitNoop(comment);
            emitPush(RA, "backup my return address.");
            emitPush(FP, "backup parent fp.");
            backend.emitADDI(FP, SP, 2 * backend.getWordSize(), "Change FP to apply new frame.");

            backend.emitJ(new Label("$" + funcName + "$saveReg"), "Jump to save callee-saved registers");
            backend.emitLocalLabel(new Label("$" + funcName + "$saveRegContinue"), "Begin real function codes:");
        }
        public void emitFunctionEnd(String funcName, registerManager regMgr, String comment) {
            // Return value should be in A0
            List<RiscVBackend.Register> savedRegs = regMgr.registerToSaveAndRestoreInFunc;

            emitNoop(comment);
            backend.emitADDI(SP, FP, -(2+savedRegs.size()) * backend.getWordSize(), "Revert all local variables on this dying frame.");
            for (int i = savedRegs.size(); i --> 0; ) {
                emitPop(savedRegs.get(i), "restore callee-saved reg.");
            }
            emitPop(FP, "restore parent FP.");
            emitPop(RA, "restore my return address.");
            backend.emitJR(RA, "Real return!");

            backend.emitLocalLabel(new Label("$" + funcName + "$saveReg"), "Save callee-saved registers before use.");
            for(int i = 0; i < savedRegs.size(); ++i) {
                emitPush(savedRegs.get(i), "save callee-saved reg.");
            }
            backend.emitJ(new Label("$" + funcName + "$saveRegContinue"), "go back to execute real function code.");
        }
        public void emitPushIntVal(RiscVBackend.Register tmpReg, Integer val, String comment) {
            emitNoop(comment);
            backend.emitLI(tmpReg, val, "INT VAL");
            emitPush(tmpReg, null);
            backend.emitLA(tmpReg, intClass.getDispatchTableLabel(), "OBJECT HEAD - DISPATCH TBL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 4, "OBJECT HEAD - SIZE = 3+1");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 1, "OBJECT HEAD - TYPE = INT");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return INT address.");
        }
        public void emitPushBoolVal(RiscVBackend.Register tmpReg, Boolean val, String comment) {
            emitNoop(comment);
            backend.emitLI(tmpReg, val ? 1 : 0, "BOOL VAL");
            emitPush(tmpReg, null);
            backend.emitLA(tmpReg, boolClass.getDispatchTableLabel(), "OBJECT HEAD - DISPATCH TBL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 4, "OBJECT HEAD - SIZE = 3+1");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 2, "OBJECT HEAD - TYPE = BOOL");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return BOOL address.");
        }
        public void emitPushStrVal(RiscVBackend.Register tmpReg, String val, String comment) {
            emitNoop(comment);
            byte [] bytes = val.getBytes(StandardCharsets.US_ASCII);
            int dataLen = bytes.length / backend.getWordSize() + 1; // Equals to ceil((len+1) DIVIDE WORD_SIZE)

            for(int cter = 0; cter < dataLen; ++cter) {
                int myRangeBegin = (dataLen-cter-1) * backend.getWordSize();
                int myRangeEnd = Math.min(myRangeBegin + backend.getWordSize(), bytes.length);
                int curr = 0;
                assert backend.getWordSize() == 4; // curr should be 4 byte = int
                String debug_pushed_str = "";
                for(int shift = 0; shift < myRangeEnd - myRangeBegin; ++shift) {
                    curr += bytes[myRangeBegin + shift] << (shift * 8);
                    debug_pushed_str += Integer.toString(bytes[myRangeBegin + shift]) + " ";
                }

                backend.emitLI(tmpReg, curr, "STR VAL - " + debug_pushed_str);
                emitPush(tmpReg, "STR VAL - PUSH A BLOCK");
            }

            backend.emitLI(tmpReg, bytes.length, "STR ATTR - __len__");
            emitPush(tmpReg, null);
            backend.emitLA(tmpReg, strClass.getDispatchTableLabel(), "OBJECT HEAD - DISPATCH TBL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3 + dataLen + 1, "OBJECT HEAD - SIZE");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3, "OBJECT HEAD - TYPE = STR");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return STR OBJ address.");
        }

    }

    // Every function needs a register manager.
    // Every variable should borrow a register, and maybe it's ok to never return it.
    //     because our toy compiler will only be tested on naive program.
    //     manager won't swap register to stack smartly (with graph coloring technic),
    //     because it doesn't worth my effort(nobody pay for that).
    //
    // FIXME:
    // Update: Every function should have a register manager.
    //         before calling another function, the register manager should determine what register is used by `another function`,
    //         and backup them before calling the function.
    //         However, if the function is recursive, then the `another function` is not analyzed yet,
    //         then there's no way to know what register is in use in the `another function`.
    //
    //         For a simple expr: `funcA() + funcB()`. If result of left operand is saved to T0, then in funcB, T0 is modified!
    //           that's bad. PLease fix the problem
    //         Maybe: result of funcA should be saved into callee-saved register S0-S8. Every function should save
    //           every callee-saved register on stack before using it.
    //                result of funcB (right operand) should be saved into caller-saved register T* or A*.
    //
    //         However, If a function save a register S0 before using it, we write `push S0`.
    //           When you want to restore S0 (before return), the stack has already messed up!!!
    //           We should save S0 to somewhere. Usually, we can do the following:
    //
    //                --- Push arguments... push arg1 ; push arg2 ; ....... ---
    //                --- I'm called! ---
    //                --- save FP ---
    //                --- I know I will use S0 and S1, save them to stack now... WAIT! HOW DO I KNOW I WILL USE S0 and S1??? --- FIXME
    //                --- I saw a expression, it uses S0... ---
    //                --- I saw a expression, it uses S1... ---
    //                --- Call others... ---
    //                --- SP(stack pointer) is messed up... I set it to FP+$(CorrectVal) ---
    //                --- pop S0 ; pop S1 ---
    //                --- restore FP, return ---
    private class registerManager {
        private Map<RiscVBackend.Register, Boolean> callerSavedUsageMap; // true if the reg is in use.
        private Map<RiscVBackend.Register, Boolean> calleeSavedUsageMap; // true if the reg is in use.
        public List<RiscVBackend.Register> registerToSaveAndRestoreInFunc;
        registerManager() {
            callerSavedUsageMap = new LinkedHashMap<>();
            calleeSavedUsageMap = new LinkedHashMap<>();

            callerSavedUsageMap.put(T0, false);
            callerSavedUsageMap.put(T1, false);
            callerSavedUsageMap.put(T2, false);
            callerSavedUsageMap.put(T3, false);
            callerSavedUsageMap.put(T4, false);
            callerSavedUsageMap.put(T5, false);
            callerSavedUsageMap.put(T6, false);
            callerSavedUsageMap.put(A1, false);
            callerSavedUsageMap.put(A2, false);
            callerSavedUsageMap.put(A3, false);
            callerSavedUsageMap.put(A4, false);
            callerSavedUsageMap.put(A5, false);
            callerSavedUsageMap.put(A6, false);
            callerSavedUsageMap.put(A7, false);
            // callerSavedUsageMap.put(A0, false); // used as return. not managed

            calleeSavedUsageMap.put(S1, false);
            calleeSavedUsageMap.put(S2, false);
            calleeSavedUsageMap.put(S3, false);
            calleeSavedUsageMap.put(S4, false);
            calleeSavedUsageMap.put(S5, false);
            calleeSavedUsageMap.put(S6, false);
            calleeSavedUsageMap.put(S7, false);
            calleeSavedUsageMap.put(S8, false);
            calleeSavedUsageMap.put(S9, false);

            registerToSaveAndRestoreInFunc = new ArrayList<>();
        }

        // For converience: we call caller-saved register as `temp register`,
        //                     call callee-saved register as `persist register` or `saved register`, because they should be saved before the function.
        public RiscVBackend.Register borrowOneTmp() {
            for(RiscVBackend.Register reg : callerSavedUsageMap.keySet()) {
                if(callerSavedUsageMap.put(reg, true) == false) {
                    return reg;
                }
            }
            throw new RuntimeException("Too many variable, not enough register.");
        }

        public RiscVBackend.Register borrowOnePersist() {
            for(RiscVBackend.Register reg : calleeSavedUsageMap.keySet()) {
                if(calleeSavedUsageMap.put(reg, true) == false) {
                    if(!registerToSaveAndRestoreInFunc.contains(reg))
                        registerToSaveAndRestoreInFunc.add(reg);
                    return reg;
                }
            }
            throw new RuntimeException("Too many variable, not enough register.");
        }

        public void returnOne(RiscVBackend.Register reg) {
            if(calleeSavedUsageMap.containsKey(reg))
                assert calleeSavedUsageMap.put(reg, false);
            else
                assert callerSavedUsageMap.put(reg, false);
        }

        /*
        public void generateCodeToBackupReg() {
            for(RiscVBackend.Register reg : registerToSaveAndRestoreInFunc) {
                betterBackend.emitPush(reg, "Backup used callee saved register in this func.");
            }
        }
        public void generateCodeToRestoreReg() {
            for(RiscVBackend.Register reg : registerToSaveAndRestoreInFunc) {
                betterBackend.emitPop(reg, "Restore backuped callee saved registers.");
            }
        }
        */
    }

    /** A code generator emitting instructions to BACKEND. */
    public CodeGenImpl(RiscVBackend backend) {
        super(backend);
        betterBackend = new BetterRsicVBackend(backend);
    }

    private BetterRsicVBackend betterBackend;

    /** Operation on None. */
    private final Label errorNone = new Label("error.None");
    /** Division by zero. */
    private final Label errorDiv = new Label("error.Div");
    /** Index out of bounds. */
    private final Label errorOob = new Label("error.OOB");

    /**
     * Emits the top level of the program.
     *
     * This method is invoked exactly once, and is surrounded
     * by some boilerplate code that: (1) initializes the heap
     * before the top-level begins and (2) exits after the top-level
     * ends.
     *
     * You only need to generate code for statements.
     *
     * @param statements top level statements
     */
    protected void emitTopLevel(List<Stmt> statements) {
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(null);
        backend.emitADDI(SP, SP, -2 * backend.getWordSize(),
                         "Saved FP and saved RA (unused at top level).");
        backend.emitSW(ZERO, SP, 0, "Top saved FP is 0.");
        backend.emitSW(ZERO, SP, 4, "Top saved RA is 0.");
        backend.emitADDI(FP, SP, 2 * backend.getWordSize(),
                         "Set FP to previous SP.");

        for (Stmt stmt : statements) {
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitLI(A0, EXIT_ECALL, "Code for ecall: exit");
        backend.emitEcall(null);
    }

    /**
     * Emits the code for a function described by FUNCINFO.
     *
     * This method is invoked once per function and method definition.
     * At the code generation stage, nested functions are emitted as
     * separate functions of their own. So if function `bar` is nested within
     * function `foo`, you only emit `foo`'s code for `foo` and only emit
     * `bar`'s code for `bar`.
     */
    protected void emitUserDefinedFunction(FuncInfo funcInfo) {
        backend.emitGlobalLabel(funcInfo.getCodeLabel());
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(funcInfo);

        betterBackend.emitFunctionBegin(funcInfo.getFuncName(), "BEGIN FUNCTION");

        for (Stmt stmt : funcInfo.getStatements()) {
            stmt.dispatch(stmtAnalyzer);
        }

        backend.emitMV(A0, ZERO, "Returning None implicitly");
        betterBackend.emitFunctionEnd(funcInfo.getFuncName(), stmtAnalyzer.regMgr, "BEGIN FUNCTION");
        backend.emitLocalLabel(stmtAnalyzer.epilogue, "Epilogue");

        // FIXME: {... reset fp etc. ...}
        backend.emitJR(RA, "Return to caller");
    }

    /** An analyzer that encapsulates code generation for statments. */
    private class StmtAnalyzer extends AbstractNodeAnalyzer<RiscVBackend.Register> {
        /*
         * The symbol table has all the info you need to determine
         * what a given identifier 'x' in the current scope is. You can
         * use it as follows:
         *   SymbolInfo x = sym.get("x");
         *
         * A SymbolInfo can be one the following:
         * - ClassInfo: a descriptor for classes
         * - FuncInfo: a descriptor for functions/methods
         * - AttrInfo: a descriptor for attributes
         * - GlobalVarInfo: a descriptor for global variables
         * - StackVarInfo: a descriptor for variables allocated on the stack,
         *      such as locals and parameters
         *
         * Since the input program is assumed to be semantically
         * valid and well-typed at this stage, you can always assume that
         * the symbol table contains valid information. For example, in
         * an expression `foo()` you KNOW that sym.get("foo") will either be
         * a FuncInfo or ClassInfo, but not any of the other infos
         * and never null.
         *
         * The symbol table in funcInfo has already been populated in
         * the base class: CodeGenBase. You do not need to add anything to
         * the symbol table. Simply query it with an identifier name to
         * get a descriptor for a function, class, variable, etc.
         *
         * The symbol table also maps nonlocal and global vars, so you
         * only need to lookup one symbol table and it will fetch the
         * appropriate info for the var that is currently in scope.
         */

        /** Symbol table for my statements. */
        private SymbolTable<SymbolInfo> sym;

        /** Label of code that exits from procedure. */
        protected Label epilogue;

        /** The descriptor for the current function, or null at the top
         *  level. */
        private FuncInfo funcInfo;

        private registerManager regMgr = new registerManager();

        /** An analyzer for the function described by FUNCINFO0, which is null
         *  for the top level. */
        StmtAnalyzer(FuncInfo funcInfo0) {
            funcInfo = funcInfo0;
            if (funcInfo == null) {
                sym = globalSymbols;
            } else {
                sym = funcInfo.getSymbolTable();
            }
            epilogue = generateLocalLabel();
        }

        // FIXME: Example of statement.
        @Override
        public RiscVBackend.Register analyze(ReturnStmt node) {
            // FIXME: Here, we emit an instruction that does nothing. Clearly,
            // this is wrong, and you'll have to fix it.
            // This is here just to demonstrate how to emit a
            // RISC-V instruction.
            backend.emitMV(ZERO, ZERO, "No-op");
            return null;
        }

        // FIXME: More, of course.
        @Override
        public RiscVBackend.Register analyze(ExprStmt node) {
            node.expr.dispatch(this);
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(CallExpr node) {
            SymbolInfo called = sym.get(node.function.name);
            if(called instanceof FuncInfo) {
                FuncInfo func = (FuncInfo) called;
                // TODO: Push the FUCKING arg0 (outter function frame ptr)
                List<RiscVBackend.Register> args_reg = new ArrayList<>();
                for(Expr expr : node.args) {
                    RiscVBackend.Register result = expr.dispatch(this);
                    if(result == null) {
                        throw new RuntimeException("NotImplemented: Expression " + expr.getClass().getName() + " returns null register.");
                    }
                    args_reg.add(result);
                }
                for(RiscVBackend.Register reg : args_reg) {
                    betterBackend.emitPush(reg, "Push function arguments.");
                    regMgr.returnOne(reg);
                }

                betterBackend.emitCall(func.getCodeLabel(), null);

                for(Object reg : args_reg)
                    betterBackend.emitPop(null, "Pop function arguments.");

                return A0; // Function return value is always in A0.
            }
            else if(called instanceof ClassInfo) {
                // TODO: Fuck the fucking framework.
            }
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(BinaryExpr node) {
            RiscVBackend.Register leftReturnReg = node.left.dispatch(this);
            if(leftReturnReg == null) return null;
            RiscVBackend.Register leftRes = regMgr.borrowOnePersist();
            backend.emitMV(leftRes, leftReturnReg, "Move left operand result to a callee saved register");
            regMgr.returnOne(leftReturnReg);
            // because right operand may overwrite temp registers.
            RiscVBackend.Register rightReturnReg = node.right.dispatch(this);
            RiscVBackend.Register rightRes = rightReturnReg;
            if(rightReturnReg == null)
                return null;
            switch(node.operator) {
                case "+":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOneTmp();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator+ Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator+ Fetch right int result");
                        backend.emitADD(leftRes, leftRes, rightRes, "Operator+::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator+::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
                    if(node.left.getInferredType().isListType()) {
                        // TODO: Merge two list.
                        throw new RuntimeException("NOTIMPLEMENTED: ListMerge");
                    }
                    if(node.left.getInferredType().equals(SymbolType.STR_TYPE)) {
                        // TODO: Merge two str.
                        throw new RuntimeException("NOTIMPLEMENTED: StrMerge");
                    }
                case "-":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOneTmp();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator- Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator- Fetch right int result");
                        backend.emitSUB(leftRes, leftRes, rightRes, "Operator-::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator-::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
                case "*":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOneTmp();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator* Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator* Fetch right int result");
                        backend.emitMUL(leftRes, leftRes, rightRes, "Operator*::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator*::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
            }
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(Identifier node) {
            SymbolInfo id = sym.get(node.name);
            if(id instanceof GlobalVarInfo) {
                GlobalVarInfo globalVarInfo = (GlobalVarInfo) id;
                RiscVBackend.Register tmpReg = regMgr.borrowOneTmp();
                backend.emitLA(tmpReg, globalVarInfo.getLabel(), "Load address of the global var.");
                return tmpReg;
            }
            // FIXME: not implemented
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(IntegerLiteral node) {
            // emitConstant(node, ValueType.INT_TYPE, "Set constant int literal.");
            RiscVBackend.Register tmpReg = regMgr.borrowOneTmp();
            betterBackend.emitPushIntVal(tmpReg, node.value, "Push int literal");
            return tmpReg;
        }

        @Override
        public RiscVBackend.Register analyze(BooleanLiteral node) {
            RiscVBackend.Register tmpReg = regMgr.borrowOneTmp();
            betterBackend.emitPushBoolVal(tmpReg, node.value, "Push bool literal");
            return tmpReg;
        }

        @Override
        public RiscVBackend.Register analyze(StringLiteral node) {
            RiscVBackend.Register tmpReg = regMgr.borrowOneTmp();
            betterBackend.emitPushStrVal(tmpReg, node.value, "push string literal.");
            return tmpReg;
        }
    }

    /**
     * Emits custom code in the CODE segment.
     *
     * This method is called after emitting the top level and the
     * function bodies for each function.
     *
     * You can use this method to emit anything you want outside of the
     * top level or functions, e.g. custom routines that you may want to
     * call from within your code to do common tasks. This is not strictly
     * needed. You might not modify this at all and still complete
     * the assignment.
     *
     * To start you off, here is an implementation of three routines that
     * will be commonly needed from within the code you will generate
     * for statements.
     *
     * The routines are error handlers for operations on None, index out
     * of bounds, and division by zero. They never return to their caller.
     * Just jump to one of these routines to throw an error and
     * exit the program. For example, to throw an OOB error:
     *   backend.emitJ(errorOob, "Go to out-of-bounds error and abort");
     *
     */
    protected void emitCustomCode() {
        emitErrorFunc(errorNone, "Operation on None");
        emitErrorFunc(errorDiv, "Divison by zero");
        emitErrorFunc(errorOob, "Index out of bounds");
    }

    /** Emit an error routine labeled ERRLABEL that aborts with message MSG. */
    private void emitErrorFunc(Label errLabel, String msg) {
        backend.emitGlobalLabel(errLabel);
        backend.emitLI(A0, ERROR_NONE, "Exit code for: " + msg);
        backend.emitLA(A1, constants.getStrConstant(msg),
                       "Load error message as str");
        backend.emitADDI(A1, A1, getAttrOffset(strClass, "__str__"),
                         "Load address of attribute __str__");
        backend.emitJ(abortLabel, "Abort");
    }
}
